// $SBND: appdev/SRI/ParamShopCart/ConnectionUtil/Connection.cs,v 1.8 2003/06/24 10:40:37 roam Exp $

using System;
using System.Data;
using System.Data.SqlClient;
using System.Text;

using Ringlet.Misc.VCS;

namespace Ringlet.Data.ConnectionUtil
{
	/// <summary>
	/// An SQL connection abstraction.
	/// </summary>
	public class Connection : IVersioned
	{
		/// <summary>
		/// The SQL connection properties.
		/// </summary>
		protected ConnectionProperties prop;
		/// <summary>
		/// The actual SQL database connection.
		/// </summary>
		protected IDbConnection conn;

		/// <summary>
		/// Default constructor
		/// </summary>
		public Connection()
		{
			prop = null;
			conn = null;
		}

		/// <summary>
		/// Preferred constructor.
		/// </summary>
		/// <param name="nprop">the connection properties to use</param>
		public Connection(ConnectionProperties nprop)
		{
			prop = new ConnectionProperties();
			nprop.Copy(prop);
			conn = null;
		}

		/// <summary>
		/// The version control system tag.
		/// </summary>
		public virtual string VCSTag
		{
			get
			{
				return ("$Ringlet: cs/misc/Scramble/ConnectionUtil/Connection.cs,v 1.5 2003/07/09 08:37:15 roam Exp $");
			}
		}

		/// <summary>
		/// Create a database connection with a previously-specified
		/// set of properties.
		/// </summary>
		/// <returns>The newly-created database connection.</returns>
		public IDbConnection Create()
		{
			// MS SQL Server only
			StringBuilder sb;
			
			sb = new StringBuilder();
			sb.AppendFormat("Server={0}; Database={1}; User ID={2}; Password={3}",
			    prop.Server, prop.Database,
			    prop.Username, prop.Password);
			conn = new SqlConnection(sb.ToString());
			return (conn);
		}

		/// <summary>
		/// Create a database connection, specifying the connection
		/// properties.
		/// </summary>
		/// <param name="nprop">the connection properties to use</param>
		/// <returns>The newly-created database connection.</returns>
		public IDbConnection Create(ConnectionProperties nprop)
		{
			nprop.Copy(prop);
			return (Create());
		}

		/// <summary>
		/// Test a database connection parameters.
		/// </summary>
		/// <returns>True if the connection is successful.</returns>
		public bool Test()
		{
			IDbConnection conn;

			conn = Create();
			conn.Open();
			conn.Close();
			return (true);
		}

		/// <summary>
		/// Open a connection to the SQL database.
		/// </summary>
		public void Open()
		{

			if (conn != null && conn.State == ConnectionState.Open)
				return;
			if (conn == null)
				Create();
			conn.Open();
		}

		/// <summary>
		/// Close the connection to the SQL database.
		/// </summary>
		public void Close()
		{

			if (conn == null)
				return;
			conn.Close();
		}

		/// <summary>
		/// Open a connection if it is not already open.
		/// </summary>
		/// <returns>The connection state at time of invocation.</returns>
		public ConnectionState StateOpen()
		{
			ConnectionState state;

			if (conn == null)
				Create();
			state = conn.State;
			if (state != ConnectionState.Open)
				Open();
			return (state);
		}

		/// <summary>
		/// Restore a previously saved connection state.
		/// </summary>
		/// <param name="state">the state to restore</param>
		/// <returns>The connection state at time of invocation.</returns>
		public ConnectionState StateRestore(ConnectionState state)
		{
			ConnectionState old;

			if (conn == null)
			{
				if (state == ConnectionState.Closed)
					return (state);
				Create();
			}
			old = conn.State;
			// Black magic? :)  Or simply obfuscated code?
			if ((old == ConnectionState.Closed) ==
			    (state == ConnectionState.Closed))
				return (old);
			if (state == ConnectionState.Closed)
				conn.Close();
			else
				conn.Open();
			return (old);
		}

		/// <summary>
		/// Get the actual SQL database connection.
		/// </summary>
		public IDbConnection DBConnection
		{
			get
			{
				return (conn);
			}
		}

		/// <summary>
		/// Wrap the real database connections' CreateCommand method.
		/// </summary>
		/// <returns>The newly-created database command.</returns>
		public IDbCommand CreateCommand()
		{

			return (conn.CreateCommand());
		}

		/// <summary>
		/// Obtain the ID of the last inserted record.
		/// </summary>
		/// <returns>The record ID on success.</returns>
		public int LastInsertID()
		{
			ConnectionState st;
			IDbCommand cmd;
			int res;

			// FIXME: SQL server specific...
			res = 0;
			st = ConnectionState.Closed;
			try
			{
				cmd = conn.CreateCommand();
				cmd.CommandText = "SELECT @@Identity";

				st = StateOpen();
				res = Decimal.ToInt32(
				    (decimal)cmd.ExecuteScalar());
			}
			finally
			{
				StateRestore(st);
			}
			return (res);
		}
	}
}
