using System;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Web;
using System.Web.SessionState;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.Web.UI.HtmlControls;

using Ringlet.Data.ConnectionUtil;
using Ringlet.Drawing.Scramble;
using Ringlet.Misc.VCS;

namespace ScrambleWebApp
{
	/// <summary>
	/// Display and solve (with the user's help) a puzzle.
	/// </summary>
	public class Puzzle : System.Web.UI.Page, IVersioned
	{
		/// <summary>
		/// The request analysis object.
		/// </summary>
		protected ScWebUtil scutil;
		/// <summary>
		/// The SQL connection properties.
		/// </summary>
		protected ConnectionProperties prop;
		/// <summary>
		/// The current puzzle's database ID.
		/// </summary>
		protected int pzid;
		/// <summary>
		/// The error message label.
		/// </summary>
		protected System.Web.UI.WebControls.Label lblError;
		/// <summary>
		/// The placeholder for the puzzle table.
		/// </summary>
		protected System.Web.UI.WebControls.PlaceHolder plcTable;
		/// <summary>
		/// The Quit button.
		/// </summary>
		protected System.Web.UI.WebControls.Button btnQuit;
		/// <summary>
		/// The moves display label.
		/// </summary>
		protected System.Web.UI.WebControls.Label lblMoves;
		/// <summary>
		/// The puzzle itself.
		/// </summary>
		protected SCPuzzle pz;
		/// <summary>
		/// Has the puzzle been already solved?
		/// </summary>
		protected bool solved;

		/// <summary>
		/// Initialize the ASP.NET page object.
		/// </summary>
		/// <param name="sender">Sender control (unused)</param>
		/// <param name="e">Event arguments (unused)</param>
		private void Page_Load(object sender, System.EventArgs e)
		{
			scutil = new ScWebUtil(Request);
			prop = scutil.ConnectionProperties;

			lblError.Text = "";
			pzid = -1;
			pz = null;
			solved = false;
			if (Session["newgame"] != null &&
			    (bool)Session["newgame"])
				CreateNew();
			else if (!IsPostBack && Request.Params["move"] != null)
				Move();
			DisplayPuzzle();
		}

		/// <summary>
		/// Get the version control system tag.
		/// </summary>
		public virtual string VCSTag
		{
			get
			{
				return ("$Ringlet: cs/misc/Scramble/Scramble/Puzzle.aspx.cs,v 1.2 2003/07/11 18:36:31 roam Exp $");
			}
		}

		/// <summary>
		/// Create a new puzzle.
		/// </summary>
		protected void CreateNew()
		{
			int imgid;
			
			try
			{
				imgid = (int)Session["pic_id"];
				pz = new SCPuzzle();
				pz.ConnectionProperties = prop;
				pzid = pz.CreatePuzzle(imgid, Session.SessionID);
				do
				{
					pz.Scramble(
					    (int)(pz.Image.Image.Rows *
					    pz.Image.Image.Columns * 3 + 1));
				} while (pz.IsOrdered());
				Session.Remove("newgame");
				Session["pz"] = pzid;
			}
			catch (Exception e)
			{
				lblError.Text = "Error: " + e.Message +
				    "<br />\n" + e.StackTrace.Replace("\n", "<br />\n");
			}
		}

		/// <summary>
		/// Load and display a puzzle.
		/// </summary>
		protected void DisplayPuzzle()
		{
			HtmlTable tbl;
			HtmlTableRow row;
			HtmlTableCell cell;
			HtmlAnchor a;
			HtmlImage img;
			uint x, y;
			SCDir dir;
			
			try
			{
				if (pz == null)
				{
					pzid = (int)Session["pz"];
					pz = new SCPuzzle();
					pz.ConnectionProperties = prop;
					pz.LoadPuzzle(pzid);
				}
	
				tbl = new HtmlTable();
				for (y = 0; y < pz.Image.Image.Rows; y++)
				{
					row = new HtmlTableRow();
					for (x = 0; x < pz.Image.Image.Columns; x++)
					{
						img = new HtmlImage();
						img.Src = String.Format(
						    "PuzzleFrag.aspx?pzid={0}&x={1}&y={2}{3}",
						    pzid, x, y,
						    ((!solved && x == pz.X && y == pz.Y)? "&blank=1": ""));
						img.Alt = String.Format(
						    "Puzzle {0}, ({1},{2})",
						    pzid, x, y);
						cell = new HtmlTableCell();
						a = new HtmlAnchor();
						dir = SCDir.Up;
						if (x == pz.X && y == pz.Y - 1)
							dir = SCDir.Up;
						else if (x == pz.X && y == pz.Y + 1)
							dir = SCDir.Down;
						else if (x == pz.X - 1 && y == pz.Y)
							dir = SCDir.Left;
						else if (x == pz.X + 1 && y == pz.Y)
							dir = SCDir.Right;
						else
							a = null;
						if (a != null && !solved)
						{
							a.Attributes.Add("href",
							    String.Format("Puzzle.aspx?move=1&pzid={0}&dir={1}",
							    pzid, dir.ToString()));
							a.Controls.Add(img);
							cell.Controls.Add(a);
						}
						else
						{
							cell.Controls.Add(img);
						}
						row.Cells.Add(cell);
					}
					tbl.Rows.Add(row);
				}
				plcTable.Controls.Add(tbl);
				lblMoves.Text = String.Format("Moves: {0}", pz.Moves);
			}
			catch (Exception e)
			{
				lblError.Text = "Error: " + e.Message +
				    "<br />\n" + e.StackTrace.Replace("\n", "<br />\n");
			}
		}

		/// <summary>
		/// Handle a user's move.
		/// </summary>
		protected void Move()
		{
			SCDir dir;
			
			try
			{
				pzid = int.Parse(Request.Params["pzid"]);
				dir = (SCDir)Enum.Parse(typeof(SCDir), Request.Params["dir"]);
				pz = new SCPuzzle();
				pz.ConnectionProperties = prop;
				pz.LoadPuzzle(pzid);
				if (!pz.CanMove(dir)) 
				{
					lblError.Text = "Cannot move " + dir.ToString();
					return;
				}
				pz.Move(dir);
				pz.Moves++;
				if (pz.IsOrdered())
					PuzzleSolved();
			}
			catch (Exception e)
			{
				lblError.Text = "Error: " + e.Message +
					"<br />\n" + e.StackTrace.Replace("\n", "<br />\n");
			}
		}

		/// <summary>
		/// Redirect to the HighScores screen upon solving.
		/// </summary>
		protected void PuzzleSolved()
		{

			Session["sc_rows"] = (int)pz.Image.Image.Rows;
			Session["sc_cols"] = (int)pz.Image.Image.Columns;
			Session["sc_top"] = 10;
			Session["sc_picture"] = pz.Image.Image.Name;
			Session["sc_moves"] = pz.Moves;
			Session["sc_showpic"] = pz.Image.Image.FileName;
			Response.Redirect("HighScores.aspx");
		}

		#region Web Form Designer generated code
		/// <summary>
		/// System initialization of the ASP.NET page object.
		/// </summary>
		/// <param name="e">Event arguments</param>
		override protected void OnInit(EventArgs e)
		{
			//
			// CODEGEN: This call is required by the ASP.NET Web Form Designer.
			//
			InitializeComponent();
			base.OnInit(e);
		}
		
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{    
			this.btnQuit.Click += new System.EventHandler(this.btnQuit_Click);
			this.Load += new System.EventHandler(this.Page_Load);

		}
		#endregion

		/// <summary>
		/// Handle the Quit button: redirect back to Default.aspx.
		/// </summary>
		/// <param name="sender">Sender control (unused)</param>
		/// <param name="e">Event arguments (unused)</param>
		private void btnQuit_Click(object sender, System.EventArgs e)
		{
			Session.Remove("pz");
			Session.Remove("newgame");
			Response.Redirect("Default.aspx");
		}
	}
}
