using System;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Web;
using System.Web.Services;

using Ringlet.Data.ConnectionUtil;
using Ringlet.Drawing.Scramble;
using Ringlet.Misc.VCS;

namespace ScrambleWebApp
{
	/// <summary>
	/// The webservice providing access to the puzzles and highscores
	/// to the Unscramble! Windows application (and possibly others :)
	/// </summary>
	[WebService(Namespace="http://schemas.ringlet.net/Scramble/2003/1.0/")]
	public class PuzzleData : System.Web.Services.WebService, IVersioned
	{
		/// <summary>
		/// The request analyzer object.
		/// </summary>
		protected ScWebUtil scutil;
		/// <summary>
		/// The SQL connection properties.
		/// </summary>
		protected ConnectionProperties prop;

		/// <summary>
		/// Initialize a PuzzleData object with default values.
		/// </summary>
		public PuzzleData()
		{
			//CODEGEN: This call is required by the ASP.NET Web Services Designer
			InitializeComponent();
			scutil = new ScWebUtil();
			prop = scutil.ConnectionProperties;
		}

		/// <summary>
		/// Get the version system control tag.
		/// </summary>
		public virtual string VCSTag
		{
			get
			{
				return ("$Ringlet: cs/misc/Scramble/Scramble/PuzzleData.asmx.cs,v 1.2 2003/07/12 01:55:39 roam Exp $");
			}
		}

		/// <summary>
		/// Get the version system control tag (this needs to be
		/// a method so it can be exported via a web service.. gack)
		/// </summary>
		[WebMethod]
		public virtual string GetVCSTag()
		{
			return (VCSTag);
		}

		/// <summary>
		/// Obtain a list of the available images.
		/// </summary>
		/// <returns>An array of image data structures.</returns>
		[WebMethod]
		public virtual SCImageData[] ListImages()
		{
			SCDatabase db;
			
			db = new SCDatabase();
			db.ConnectionProperties = prop;
			return (db.ListAllImages());
		}

		/// <summary>
		/// Get the byte representation of an image's thumbnail.
		/// </summary>
		/// <param name="id">The image to query.</param>
		/// <returns>A byte array containing the image data.</returns>
		[WebMethod]
		public virtual byte[] GetImageThumbnail(int id)
		{
			SCDBImage img;
			Image thumb;
			MemoryStream m;
			byte[] buf;

			img = new SCDBImage();			
			img.ConnectionProperties = prop;
			img.LoadImage(id);
			thumb = new Bitmap(img.Image.Image, 100, 100);
			m = new MemoryStream();
			thumb.Save(m, ImageFormat.Jpeg);
			buf = m.GetBuffer();
			m.Close();
			return (buf);
		}

		/// <summary>
		/// Get the byte representation of an image fragment.
		/// </summary>
		/// <param name="id">The image to query.</param>
		/// <param name="x">The fragment's X coordinate.</param>
		/// <param name="y">The fragment's Y coordinate.</param>
		/// <returns>A byte array containing the image data.</returns>
		[WebMethod]
		public virtual byte[] GetImageFragment(int id, int x, int y)
		{
			SCDBImage img;
			MemoryStream m;
			byte[] buf;

			img = new SCDBImage();
			img.ConnectionProperties = prop;
			img.LoadImage(id);
			m = new MemoryStream();
			img.Image.Pieces[y, x].Save(m, ImageFormat.Jpeg);
			buf = m.GetBuffer();
			m.Close();
			return (buf);
		}

		/// <summary>
		/// Obtain a list of all the recorded highscores.
		/// </summary>
		/// <returns>A list of all highscores.</returns>
		[WebMethod]
		public SCScore[] ListAllScores()
		{
			SCHighScores sc;
			
			sc = new SCHighScores();
			sc.ConnectionProperties = prop;
			return (sc.ListAllScores());
		}
		
		/// <summary>
		/// Obtain a list of the highscores for a given cut.
		/// </summary>
		/// <param name="rows">The cut rows.</param>
		/// <param name="cols">The cut columns.</param>
		/// <returns>The scores list for the given cut.</returns>
		[WebMethod]
		public SCScore[] ListScores(int rows, int cols)
		{
			SCHighScores sc;
			
			sc = new SCHighScores();
			sc.ConnectionProperties = prop;
			return (sc.ListAllScores(rows, cols));
		}

		/// <summary>
		/// Add a high score entry into the table.
		/// </summary>
		/// <param name="s">The high score entry to add.</param>
		/// <returns>The database ID of the newly-added entry.</returns>
		[WebMethod]
		public int AddScore(SCScore s)
		{
			Connection conn;

			conn = new Connection(prop);
			conn.Create();
			s.AddScore(conn);
			return (s.id);
		}
		
		/// <summary>
		/// Delete a high-score entry from the table.
		/// </summary>
		/// <param name="s">The entry to delete.</param>
		[WebMethod]
		public void DeleteScore(SCScore s)
		{
			Connection conn;

			conn = new Connection(prop);
			conn.Create();
			s.DeleteScore(conn);
		}

		#region Component Designer generated code
		
		//Required by the Web Services Designer 
		private IContainer components = null;
				
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if(disposing && components != null)
			{
				components.Dispose();
			}
			base.Dispose(disposing);		
		}
		
		#endregion

		// WEB SERVICE EXAMPLE
		// The HelloWorld() example service returns the string Hello World
		// To build, uncomment the following lines then save and build the project
		// To test this web service, press F5

//		[WebMethod]
//		public string HelloWorld()
//		{
//			return "Hello World";
//		}
	}
}
