using System;
using System.Data;
using System.Drawing;
using System.IO;
using System.Text;

using Ringlet.Data.ConnectionUtil;
using Ringlet.Misc.VCS;

namespace Ringlet.Drawing.Scramble
{
	/// <summary>
	/// Summary description for SCDBImage.
	/// </summary>
	public class SCDBImage : IVersioned
	{
		/// <summary>
		/// The cut-up image object.
		/// </summary>
		protected SCImage image;
		/// <summary>
		/// The SQL connection properties;
		/// </summary>
		protected ConnectionProperties prop;
		/// <summary>
		/// The SQL connection itself.
		/// </summary>
		protected Connection conn;

		/// <summary>
		/// Initialize a SCDBImage object using the default values.
		/// </summary>
		public SCDBImage()
		{
			image = null;
		}

		/// <summary>
		/// The version control system tag.
		/// </summary>
		public virtual string VCSTag
		{
			get
			{
				return ("$Ringlet: cs/misc/Scramble/libscpic/SCDBImage.cs,v 1.8 2003/07/12 01:51:18 roam Exp $");
			}
		}

		/// <summary>
		/// Get or set the cut-up image object.
		/// </summary>
		public SCImage Image
		{
			get
			{
				return (image);
			}

			set
			{
				image = value;
			}
		}

		/// <summary>
		/// Get or set the SQL connection properties.
		/// </summary>
		public ConnectionProperties ConnectionProperties
		{
			get
			{
				return (prop);
			}

			set
			{
				prop = value;
				if (conn != null)
					conn.Close();
				conn = new Connection(prop);
				conn.Create();
			}
		}

		/// <summary>
		/// Store the image into the database.
		/// </summary>
		/// <returns>The image identifier in the database.</returns>
		public int StoreImage()
		{
			ConnectionState st;
			IDbCommand cmd;
			IDbDataParameter param;
			int id;
			uint x, y, cols, rows;
			StringBuilder sbcmd;

			id = 0;
			st = ConnectionState.Closed;
			try
			{
				cmd = conn.CreateCommand();
				cmd.CommandText =
					@"INSERT INTO scpics(pic_fname, pic_date, pic_slx, pic_sly, pic_inuse,
    pic_fragdir, pic_fragprefix, pic_fragsuffix, pic_name, pic_comment)
VALUES(@fname, @date, @x, @y, 0, @dir, @prefix, @suffix, @pic_name, @pic_comment)";

				param = cmd.CreateParameter();
				param.DbType = DbType.String;
				param.ParameterName = "@fname";
				param.Value = image.FileName;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.DbType = DbType.DateTime;
				param.ParameterName = "@date";
				param.Value = DateTime.Now;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.DbType = DbType.Int32;
				param.ParameterName = "@x";
				param.Value = image.Columns;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.DbType = DbType.Int32;
				param.ParameterName = "@y";
				param.Value = image.Rows;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.DbType = DbType.String;
				param.ParameterName = "@dir";
				param.Value = image.FileDir;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.DbType = DbType.String;
				param.ParameterName = "@prefix";
				param.Value = image.FilePrefix;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.DbType = DbType.String;
				param.ParameterName = "@suffix";
				param.Value = image.FileSuffix;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.DbType = DbType.String;
				param.ParameterName = "@pic_name";
				param.Value = image.Name;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.DbType = DbType.String;
				param.ParameterName = "@pic_comment";
				param.Value = image.Comment;
				cmd.Parameters.Add(param);

				st = conn.StateOpen();
				cmd.ExecuteNonQuery();

				id = conn.LastInsertID();

#if ALL_AT_ONCE
				cmd = conn.CreateCommand();
				param = cmd.CreateParameter();
				param.ParameterName = "@id";
				param.DbType = DbType.Int32;
				param.Value = id;
				cmd.Parameters.Add(param);

				sbcmd = new StringBuilder();
				sbcmd.Append(
@"INSERT INTO scfragments(frag_pic, frag_x, frag_y, frag_fname) VALUES ");

#endif // ALL_AT_ONCE
				cols = image.Columns;
				rows = image.Rows;
				for (y = 0; y < rows; y++)
					for (x = 0; x < cols; x++)
					{
#if !ALL_AT_ONCE
						cmd = conn.CreateCommand();
						param = cmd.CreateParameter();
						param.ParameterName = "@id";
						param.DbType = DbType.Int32;
						param.Value = id;
						cmd.Parameters.Add(param);
						sbcmd = new StringBuilder();
						sbcmd.Append(
@"INSERT INTO scfragments(frag_pic, frag_x, frag_y, frag_fname) VALUES ");
#endif // !ALL_AT_ONCE

						param = cmd.CreateParameter();
						param.ParameterName =
						    String.Format("@fname_{0}_{1}",
						    y, x);
						param.DbType = DbType.String;
						param.Value = image.Names[y, x];
						cmd.Parameters.Add(param);
						sbcmd.AppendFormat(
						    "(@id, {0}, {1}, {2})",
						    x, y, param.ParameterName);
#if ALL_AT_ONCE
						if (x < cols - 1 || y < rows - 1)
							sbcmd.Append(", ");
#else // ALL_AT_ONCE
						cmd.CommandText = sbcmd.ToString();
						cmd.ExecuteNonQuery();
#endif // ALL_AT_ONCE
					}
#if ALL_AT_ONCE
				cmd.CommandText = sbcmd.ToString();
				cmd.ExecuteNonQuery();
#endif // ALL_AT_ONCE
			}
			catch (Exception ex)
			{
				if (id != 0)
				{
					try
					{
						cmd = conn.CreateCommand();
						cmd.CommandText =
@"DELETE FROM scpic WHERE pic_id = @id";
						param = cmd.CreateParameter();
						param.ParameterName = "@id";
						param.DbType = DbType.Int32;
						param.Value = id;
						cmd.Parameters.Add(param);
						cmd.ExecuteNonQuery();

						cmd = conn.CreateCommand();
						cmd.CommandText =
@"DELETE FROM scfragments WHERE frag_pic = @id";
						param = cmd.CreateParameter();
						param.ParameterName = "@id";
						param.DbType = DbType.Int32;
						param.Value = id;
						cmd.Parameters.Add(param);
						cmd.ExecuteNonQuery();
					}
					catch (Exception)
					{
					}
				}
				throw (ex);
			}
			finally
			{

				conn.StateRestore(st);
			}
			return (id);
		}

		/// <summary>
		/// Load information about a fragmented image from the database.
		/// </summary>
		/// <param name="id">The ID of the image to load.</param>
		public void LoadImage(int id)
		{
			ConnectionState st;
			IDbCommand cmd;
			IDbDataParameter param;
			IDataReader rdr;
			SCImage img;
			uint rows, cols, x, y;
			System.Drawing.Image[,] pieces;
			Image dimg;
			string[,] names;
			string fname;
			Stream istr;

			st = ConnectionState.Closed;
			istr = null;
			rdr = null;
			try
			{
				cmd = conn.CreateCommand();
				cmd.CommandText =
@"SELECT * FROM scpics WHERE pic_id=@id";
				param = cmd.CreateParameter();
				param.ParameterName = "@id";
				param.DbType = DbType.Int32;
				param.Value = id;
				cmd.Parameters.Add(param);

				st = conn.StateOpen();
				rdr = cmd.ExecuteReader();
				if (rdr == null || !rdr.Read())
					throw (new ApplicationException(
					    "No image found with id " +
					    id.ToString()));

				fname = (string)rdr["pic_fname"];
				istr = new FileStream(fname, FileMode.Open);
				img = new SCImage();
				img.Image = new Bitmap(new Bitmap(istr));
				istr.Close();
				istr = null;
				
				img.FileName = fname;
				img.FileDir = (string)rdr["pic_fragdir"];
				img.FilePrefix = (string)rdr["pic_fragprefix"];
				img.FileSuffix = (string)rdr["pic_fragsuffix"];
				img.Name = (string)rdr["pic_name"];
				img.Comment = (string)rdr["pic_comment"];
				cols = (uint)(int)rdr["pic_slx"];
				rows = (uint)(int)rdr["pic_sly"];
				rdr.Close();
				rdr = null;

				pieces = new System.Drawing.Image[rows, cols];
				names = new string[rows, cols];

				cmd = conn.CreateCommand();
				cmd.CommandText =
@"SELECT * FROM scfragments WHERE frag_pic = @id";
				param = cmd.CreateParameter();
				param.ParameterName = "@id";
				param.DbType = DbType.Int32;
				param.Value = id;
				cmd.Parameters.Add(param);

				rdr = cmd.ExecuteReader();
				if (rdr == null || !rdr.Read())
					throw (new ApplicationException(
					    "No fragments found for image " +
					    id.ToString()));
				do 
				{
					x = (uint)(int)rdr["frag_x"];
					y = (uint)(int)rdr["frag_y"];
					if (x >= cols || y >= rows)
						throw (new ApplicationException(
						    String.Format(
                                                    "Invalid coordinates for image fragment {0}: ({1}, {2}) out of range for ({3}, {4})",
						    (int)rdr["frag_id"], x, y,
						    cols, rows)));
					names[y, x] = (string)rdr["frag_fname"];
					istr = new FileStream(names[y, x],
					    FileMode.Open);
					dimg = new Bitmap(new Bitmap(istr));
					pieces[y, x] = dimg;
					istr.Close();
					istr = null;
				} while (rdr.Read());
				rdr.Close();
				rdr = null;

				img.Pieces = pieces;
				img.Names = names;
				Image = img;
			}
			finally
			{
				if (istr != null)
					try
					{
						istr.Close();
					}
					catch {}
				if (rdr != null)
					try
					{
						rdr.Close();
					}
					catch {}
				try
				{
					conn.StateRestore(st);
				}
				catch {}
			}
		}
	}
}
