using System;
using System.Collections;
using System.Data;

using Ringlet.Data.ConnectionUtil;
using Ringlet.Misc.VCS;

namespace Ringlet.Drawing.Scramble
{
	/// <summary>
	/// A class for manipulation of the high scores table.
	/// </summary>
	public class SCHighScores : IVersioned
	{
		/// <summary>
		/// The SQL connection properties.
		/// </summary>
		ConnectionProperties prop;
		/// <summary>
		/// The SQL connection itself.
		/// </summary>
		Connection conn;
		SCScore[] scores;
		int rows;
		int cols;
		int num;

		/// <summary>
		/// A delegate type for adding a score to the tally.
		/// </summary>
		public delegate int AddScoreDelegate(ref SCScore s, Connection conn);
		/// <summary>
		/// A delegate type for removing a score from the tally.
		/// </summary>
		public delegate void DeleteScoreDelegate(SCScore s, Connection conn);
		/// <summary>
		/// A delegate type for listing all scores on the tally.
		/// </summary>
		public delegate SCScore[] ListAllScoresDelegate(Connection conn);
		/// <summary>
		/// The routine to use to add scores, if other than
		/// SCScore.AddScore()
		/// </summary>
		protected AddScoreDelegate dlgAdd;
		/// <summary>
		/// The routine to use to delete scores, if other than
		/// SCScore.DeleteScore()
		/// </summary>
		protected DeleteScoreDelegate dlgDelete;
		/// <summary>
		/// The routine to use to list scores, if other than
		/// ListAllScores()
		/// </summary>
		protected ListAllScoresDelegate dlgList;

		/// <summary>
		/// Initialize a SCHighScores object with default values.
		/// </summary>
		public SCHighScores()
		{
			prop = null;
			conn = null;
			scores = null;
			rows = cols = 0;
			num = 10;
		}

		/// <summary>
		/// The version control system tag.
		/// </summary>
		public virtual string VCSTag
		{
			get
			{
				return ("$Ringlet: cs/misc/Scramble/libscpic/SCHighScores.cs,v 1.2 2003/07/12 01:55:40 roam Exp $");
			}
		}

		/// <summary>
		/// Get or set the current score table's image rows.
		/// </summary>
		public int Rows
		{
			get
			{
				return (rows);
			}

			set
			{
				rows = value;
			}
		}

		/// <summary>
		/// Get or set the current score table's image columns.
		/// </summary>
		public int Columns
		{
			get
			{
				return (cols);
			}

			set
			{
				cols = value;
			}
		}

		/// <summary>
		/// Get or set the current score table's length.
		/// </summary>
		public int ScoreNum
		{
			get
			{
				return (num);
			}

			set
			{
				num = value;
			}
		}

		/// <summary>
		/// Get or set the current highscore table.
		/// </summary>
		public SCScore[] Scores
		{
			get
			{
				return (scores);
			}

			set
			{
				scores = value;
				num = scores.Length;
			}
		}

		/// <summary>
		/// Get or set the SQL connection properties.
		/// </summary>
		public ConnectionProperties ConnectionProperties
		{
			get
			{
				return (prop);
			}

			set
			{
				prop = value;
				if (conn != null)
					conn.Close();
				conn = new Connection(prop);
				conn.Create();
			}
		}

		/// <summary>
		/// Get or set the delegate for adding scores.
		/// </summary>
		public AddScoreDelegate AddScoreDlg
		{
			get
			{
				return (dlgAdd);
			}

			set
			{
				dlgAdd = value;
			}
		}

		/// <summary>
		/// Get or set the delegate for deleting scores.
		/// </summary>
		public DeleteScoreDelegate DeleteScoreDlg
		{
			get
			{
				return (dlgDelete);
			}

			set
			{
				dlgDelete = value;
			}
		}

		/// <summary>
		/// Get or set the delegate for listing scores.
		/// </summary>
		public ListAllScoresDelegate ListAllScoresDlg
		{
			get
			{
				return (dlgList);
			}

			set
			{
				dlgList = value;
			}
		}

		/// <summary>
		/// Obtain a list of all the high scores in the database.
		/// </summary>
		/// <returns>An array of high score objects.</returns>
		public SCScore[] ListAllScores()
		{
			ConnectionState st;
			IDbCommand cmd;
			IDataReader rdr;
			ArrayList l;
			SCScore s;

			if (dlgList != null)
				return (dlgList(conn));

			st = ConnectionState.Closed;
			l = new ArrayList();
			try
			{
				cmd = conn.CreateCommand();
				cmd.CommandText =
@"SELECT * FROM scscores";
				st = conn.StateOpen();
				rdr = cmd.ExecuteReader();
				if (rdr == null)
					throw (new ApplicationException(
					    "No Scramble high scores found"));
				while (rdr.Read())
				{
					s = new SCScore(rdr);
					l.Add(s);
				}
			}
			finally
			{
				try
				{
					conn.StateRestore(st);
				}
				catch {}
			}
			return (SCScore[])l.ToArray(typeof(SCScore));
		}

		/// <summary>
		/// Obtain a list of all the high scores for a specific cut.
		/// </summary>
		/// <param name="rows">The number of image rows.</param>
		/// <param name="cols">The number of image columns.</param>
		/// <returns>An array of high-score objects.</returns>
		public SCScore[] ListAllScores(int rows, int cols)
		{
			ArrayList l;
			SCScore[] all;

			l = new ArrayList();
			all = ListAllScores();
			foreach (SCScore sc in all)
				if (sc.rows == rows && sc.cols == cols)
					l.Add(sc);
			return (SCScore[])l.ToArray(typeof(SCScore));
		}

		/// <summary>
		/// Obtain a list of the top scores for a specific cut.
		/// </summary>
		/// <param name="rows">The number of image rows.</param>
		/// <param name="cols">The number of image columns.</param>
		/// <param name="num">The number of top scores to return.</param>
		/// <returns>An array of high-score objects.</returns>
		public SCScore[] ListTop(int rows, int cols, int num)
		{
			SCScore[] all, top;

			if (num < 1)
				throw (new ArgumentException(
				    "The number of top scores must be positive",
				    "num"));
			all = ListAllScores(rows, cols);
			if (all.Length == 0)
				return (all);
			Array.Sort(all);
			if (all.Length <= num)
				return (all);
			
			top = new SCScore[num];
			Array.Copy(all, all.Length - num, top, 0, num);
			Array.Reverse(top);
			return (top);
		}

		/// <summary>
		/// Store the list of top-scores for a given cut.
		/// </summary>
		/// <returns>An array of high-score objects.</returns>
		public SCScore[] SelectTop()
		{

			scores = ListTop(rows, cols, num);
			return (scores);
		}

		/// <summary>
		/// Determine whether the player has scored.
		/// </summary>
		/// <param name="moves">The number of moves completed.</param>
		/// <returns>True if a high score has been achieved.</returns>
		public bool IsHighScore(int moves)
		{

			return (scores.Length < num ||
			    moves <= scores[scores.Length - 1].moves);
		}
		
		/// <summary>
		/// Add an entry to the current score table.
		/// </summary>
		/// <param name="s">The score to add.</param>
		/// <returns>The position that the score was added at.</returns>
		public int UpdateScore(SCScore s)
		{
			int i;

			SelectTop();
			if (scores.Length > 0)
			{
				for (i = scores.Length; i > 0; i--)
					if (scores[i - 1].moves < s.moves)
						break;
				if (i == scores.Length && i >= num)
					return (scores.Length + 1);
				if (scores.Length >= num)
				{
					if (dlgDelete != null)
						dlgDelete(scores[scores.Length - 1],
						    conn);
					else
						scores[scores.Length - 1].
						    DeleteScore(conn);
				}
				else
				{
					SCScore[] newarr;

					newarr = new SCScore[scores.Length + 1];
					scores.CopyTo(newarr, 0);
					scores = newarr;
				}
			}
			else
			{
				scores = new SCScore[1];
				i = 0;
			}
			if (dlgAdd != null)
				dlgAdd(ref s, conn);
			else
				s.AddScore(conn);
			if (i < scores.Length - 1)
				Array.Copy(scores, i, scores, i + 1, scores.Length - i - 1);
			scores[i] = s;
			return (i);
		}
	}
}
