using System;
using System.Collections;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;

using Ringlet.Misc.VCS;

namespace Ringlet.Drawing.Scramble
{
	/// <summary>
	/// A class representing an image "cut up" into a puzzle grid.
	/// </summary>
	public class SCImage : IVersioned
	{
		/// <summary>
		/// The image to be "cut up".
		/// </summary>
		protected Image img;
		/// <summary>
		/// The image data structure.
		/// </summary>
		protected SCImageData data;
		/// <summary>
		/// The list of image fragments.
		/// </summary>
		protected Image[,] pieces;
		/// <summary>
		/// The list of image fragment filenames.
		/// </summary>
		protected string[,] fragnames;

		/// <summary>
		/// The VCS tag for this file.
		/// </summary>
		public virtual string VCSTag
		{
			get
			{
				return ("$Ringlet: cs/misc/Scramble/libscpic/SCImage.cs,v 1.5 2003/07/10 07:42:36 roam Exp $");
			}
		}

		/// <summary>
		/// Initialize a SCImage object with default values.
		/// </summary>
		public SCImage()
		{
			img = null;
			pieces = null;
			data = new SCImageData();
			data.id = -1;
			data.filesuffix = ".jpg";
		}

		/// <summary>
		/// Initialize a SCImage object with the supplied image.
		/// </summary>
		/// <param name="nimg">The image to use.</param>
		public SCImage(Image nimg) : this()
		{
			Image = nimg;
		}

		/// <summary>
		/// Get or set the image to be cut up.
		/// </summary>
		public Image Image
		{
			get
			{
				return (img);
			}

			set
			{
				img = value;
				data = new SCImageData();
				data.id = -1;
				data.filesuffix = ".jpg";
			}
		}

		/// <summary>
		/// Get or set the original filename of the whole image.
		/// </summary>
		public string FileName
		{
			get
			{
				return (data.filename);
			}

			set
			{
				data.filename = value;
			}
		}

		/// <summary>
		/// Get or set the image fragments' directory.
		/// </summary>
		public string FileDir
		{
			get
			{
				return (data.filedir);
			}

			set
			{
				data.filedir = value;
			}
		}

		/// <summary>
		/// Get or set the image fragments' name prefix.
		/// </summary>
		public string FilePrefix
		{
			get
			{
				return (data.fileprefix);
			}

			set
			{
				data.fileprefix = value;
			}
		}

		/// <summary>
		/// Get or set the image framgents' name suffix.
		/// </summary>
		public string FileSuffix
		{
			get
			{
				return (data.filesuffix);
			}

			set
			{
				data.filesuffix = value;
			}
		}

		/// <summary>
		/// Get the number of columns the image has been cut up into.
		/// </summary>
		public uint Columns
		{
			get
			{
				return (data.cols);
			}
		}

		/// <summary>
		/// Get the number of rows the image has been cut up into.
		/// </summary>
		public uint Rows
		{
			get
			{
				return (data.rows);
			}
		}

		/// <summary>
		/// Get the list of fragments.
		/// </summary>
		public Image[,] Pieces
		{
			get
			{
				return (pieces);
			}

			set
			{
				pieces = value;
				data.rows = (uint)(pieces.GetUpperBound(0) -
				    pieces.GetLowerBound(0) + 1);
				data.cols = (uint)(pieces.GetUpperBound(1) -
				    pieces.GetLowerBound(1) + 1);
			}
		}

		/// <summary>
		/// Get or set the list of image fragment filenames.
		/// </summary>
		public string[,] Names
		{
			get
			{
				return (fragnames);
			}

			set
			{
				if (pieces == null)
					throw (new ApplicationException(
					    "Cannot assign an array of names without an array of pieces"));
				if (data.rows !=
				    value.GetUpperBound(0) - value.GetLowerBound(0) + 1)
					throw (new ApplicationException(
					    "Row number mismatch between names and pieces arrays"));
				if (data.cols !=
				    value.GetUpperBound(1) - value.GetLowerBound(1) + 1)
					throw (new ApplicationException(
					    "Column number mismatch between names and pieces arrays"));
				fragnames = value;
			}
		}

		/// <summary>
		/// Get or set the image name.
		/// </summary>
		public string Name
		{
			get
			{
				return (data.name);
			}

			set
			{
				data.name = value;
			}
		}

		/// <summary>
		/// Get or set the image comment text.
		/// </summary>
		public string Comment
		{
			get
			{
				return (data.comment);
			}

			set
			{
				data.comment = value;
			}
		}

		/// <summary>
		/// Cut the image into the specified number of rows and
		/// columns.
		/// </summary>
		/// <param name="x">The number of columns to cut the image
		/// into.</param>
		/// <param name="y">The number of rows to cut the image
		/// into.</param>
		public void Cut(uint x, uint y)
		{
			GraphicsUnit unit;
			Image small;
			Graphics gfx;
			RectangleF size, rect;
			float lx, ly;
			int cx, cy, px, py;
			
			unit = GraphicsUnit.Pixel;
			size = img.GetBounds(ref unit);
			lx = size.Width;
			ly = size.Height;
			px = (int)(lx / x);
			py = (int)(ly / y);

			pieces = new Image[y, x];
			
			rect = new RectangleF(0, 0, 0, 0);
			for (cy = 0; cy < y; cy++)
			{
				rect.Y = cy * py;
				if (cy < y - 1)
					rect.Height = py;
				else
					rect.Height = ly - py * (y - 1);
				for (cx = 0; cx < x; cx++)
				{
					rect.X = cx * px;
					if (cx < x - 1)
						rect.Width = px;
					else
						rect.Width = lx - px * (x - 1);
					small = new Bitmap((int)rect.Width,
					    (int)rect.Height,
					    PixelFormat.Format24bppRgb);
					gfx = Graphics.FromImage(small);
					gfx.DrawImage(img, 0, 0, rect, unit);
					pieces[cy, cx] = small;
				}
			}
			data.cols = x;
			data.rows = y;
		}

		/// <summary>
		/// Save the image's pieces into a set of files.
		/// </summary>
		/// <param name="dir">The directory to save into.</param>
		/// <param name="prefix">The prefix to use for the filenames.</param>
		public void SavePieces(string dir, string prefix)
		{
			int cx, cy, idx;
			string[,] names;
			Stream ostr;
			bool success;
			
			if (dir != null)
				data.filedir = dir;
			if (prefix != null)
				data.fileprefix = prefix;

			// Here's hoping this is initialized with nulls...
			names = new string[
			    pieces.GetUpperBound(0) - pieces.GetLowerBound(0) + 1,
			    pieces.GetUpperBound(1) - pieces.GetLowerBound(1) + 1];
			
			idx = cx = cy = 0;
			ostr = null;
			success = false;
			try
			{
				for (cy = pieces.GetLowerBound(0);
				    cy <= pieces.GetUpperBound(0); cy++)
				{
					for (cx = pieces.GetLowerBound(1);
					    cx <= pieces.GetUpperBound(1);
					    cx++)
					{
						ostr = NextFilename(dir,
						    data.fileprefix,
						    data.filesuffix,
						    ref idx, out names[cy, cx]);
						if (ostr == null)
							throw (new
							    ApplicationException(
							    String.Format("Could not get a name for the piece at ({0},{1}), index {2}",
							    cx, cy, idx)));
						pieces[cy, cx].Save(ostr, ImageFormat.Jpeg);
						ostr.Close();
						ostr = null;
					}
				}
				success = true;
			}
			finally
			{
				if (ostr != null)
					try
					{
						ostr.Close();
					}
					catch (Exception)
					{}
				if (!success)
				{
					RemoveFiles(names, cx, cy);
					names = null;
				}
			}
			fragnames = names;
		}

		/// <summary>
		/// Remove the files whose names are stored in the array.
		/// </summary>
		/// <param name="names">The array of file names.</param>
		/// <param name="x">The current column.</param>
		/// <param name="y">The current row.</param>
		protected void RemoveFiles(string[,] names, int x, int y)
		{
			int dx, cx, cy;

			dx = names.GetUpperBound(1) - names.GetLowerBound(1);
			for (cy = 0; cy <= y; cy++)
				for (cx = 0; cx <= dx; cx++)
				{
					if (names[cy, cx] != null)
						try
						{
							File.Delete(names[cy, cx]);
						}
						catch (Exception)
						{}
					if (cy == y && cx == x)
						break;
				}
		}

		/// <summary>
		/// Create a new file in the specified directory, using
		/// the specified prefix.
		/// </summary>
		/// <param name="dir">The directory to create into.</param>
		/// <param name="prefix">The filename prefix to use.</param>
		/// <param name="suffix">The filename suffix to use.</param>
		/// <param name="cidx">The current index.</param>
		/// <param name="name">The output file name.</param>
		/// <returns>A stream for writing into the newly-created
		/// file on success; 0 if the directory already contains
		/// 1000 files with that prefix.</returns>
		protected Stream NextFilename(string dir, string prefix,
			string suffix, ref int cidx, out string name)
		{
			int idx;
			string fname;
			Stream s;

			for (idx = cidx + 1; idx < 1000; idx++)
			{
				fname = String.Format("{0}{1}{2:D3}{3}",
				    dir, prefix, idx, suffix);
				try
				{
					s = File.Open(fname, FileMode.CreateNew);
					cidx = idx;
					name = fname;
					return (s);
				}
				catch (Exception)
				{}
			}
			name = null;
			return (null);
		}

		/// <summary>
		/// Get or set the data structure describing the image.
		/// </summary>
		public SCImageData Data
		{
			get
			{
				return (data);
			}

			set
			{
				data = value;
				data.rows = data.cols = 0;
				pieces = null;
				fragnames = null;
			}
		}
	}
}
