using System;
using System.Collections;
using System.Data;
using System.Drawing;
using System.IO;

using Ringlet.Data.ConnectionUtil;
using Ringlet.Misc.VCS;

namespace Ringlet.Drawing.Scramble
{
	/// <summary>
	/// A puzzle: a fragmented, scrambled image.
	/// </summary>
	public class SCPuzzle : IVersioned
	{
		/// <summary>
		/// The SQL connection properties.
		/// </summary>
		ConnectionProperties prop;
		/// <summary>
		/// The SQL connection itself.
		/// </summary>
		protected Connection conn;
		/// <summary>
		/// The database idea of the puzzle's image.
		/// </summary>
		protected SCDBImage image;
		/// <summary>
		/// The location of the scrambled image pieces.
		/// </summary>
		protected SCPuzzleFrag[,] pieces;
		/// <summary>
		/// The puzzle ID.
		/// </summary>
		protected int id;
		/// <summary>
		/// The moves the player has made so far.
		/// </summary>
		protected int moves;
		/// <summary>
		/// The X coordinate of the current piece's position.
		/// </summary>
		protected int curx;
		/// <summary>
		/// The Y coordinate of the current piece's position.
		/// </summary>
		protected int cury;
		/// <summary>
		/// Flag: hold the database updates for the present.
		/// </summary>
		protected bool hold;
		/// <summary>
		/// Flag: disable any database interaction at all.
		/// </summary>
		protected bool disabledb;
		
		/// <summary>
		/// Initialize a SCPuzzle object with default values.
		/// </summary>
		public SCPuzzle()
		{
			conn = null;
			prop = null;
			image = null;
			pieces = null;
			id = curx = cury = -1;
			moves = 0;
			hold = disabledb = false;
		}

		/// <summary>
		/// Get the version control system tag.
		/// </summary>
		public virtual string VCSTag
		{
			get
			{
				return ("$Ringlet: cs/misc/Scramble/libscpic/SCPuzzle.cs,v 1.5 2003/07/11 23:32:02 roam Exp $");
			}
		}

		/// <summary>
		/// Get or set the SQL connection properties.
		/// </summary>
		public ConnectionProperties ConnectionProperties
		{
			get
			{
				return (prop);
			}

			set
			{
				if (conn != null)
					conn.Close();
				prop = value;
				conn = new Connection(prop);
				conn.Create();
			}
		}

		/// <summary>
		/// Get the database image handler for the puzzle.
		/// </summary>
		public SCDBImage Image
		{
			get
			{
				return (image);
			}
		}

		/// <summary>
		/// Get the X coordinate of the current piece.
		/// </summary>
		public int X
		{
			get
			{
				return (curx);
			}
		}

		/// <summary>
		/// Get the Y coordinate of the current piece.
		/// </summary>
		public int Y
		{
			get
			{
				return (cury);
			}
		}

		/// <summary>
		/// Create a puzzle from the specified image.
		/// </summary>
		/// <param name="imgid">The database ID of the image.</param>
		/// <param name="sess">The user's session ID.</param>
		/// <returns>The database ID of the new puzzle.</returns>
		public int CreatePuzzle(int imgid, string sess)
		{
			int x, y;
			ConnectionState st;
			IDbCommand cmd;
			IDbDataParameter param;

			if (disabledb)
				throw (new ApplicationException(
				    "For operation without a database, use BuildPuzzle()"));

			id = -1;
			st = ConnectionState.Closed;
			try
			{
				image = new SCDBImage();
				image.ConnectionProperties = prop;
				image.LoadImage(imgid);
				curx = (int)image.Image.Columns - 1;
				cury = (int)image.Image.Rows - 1;

				InitPieces();

				cmd = conn.CreateCommand();
				cmd.CommandText =
@"INSERT INTO scpuzzles(pz_pic, pz_session, pz_moves, pz_curx, pz_cury) VALUES(@pic, @sess, 0, @curx, @cury)";
				
				param = cmd.CreateParameter();
				param.ParameterName = "@pic";
				param.DbType = DbType.Int32;
				param.Value = imgid;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.ParameterName = "@sess";
				param.DbType = DbType.String;
				param.Value = sess;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.ParameterName = "@curx";
				param.DbType = DbType.Int32;
				param.Value = curx;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.ParameterName = "@cury";
				param.DbType = DbType.Int32;
				param.Value = cury;
				cmd.Parameters.Add(param);

				st = conn.StateOpen();
				cmd.ExecuteNonQuery();

				id = conn.LastInsertID();

				for (y = 0; y < image.Image.Rows; y++)
					for (x = 0; x < image.Image.Columns; x++)
					{
						cmd = conn.CreateCommand();
						cmd.CommandText = String.Format(
@"INSERT INTO scpuzzlefrag(pzf_puzzle, pzf_x, pzf_y, pzf_imx, pzf_imy)
VALUES({0}, {1}, {2}, {3}, {4})",
							id, x, y, x, y);
						cmd.ExecuteNonQuery();
					}
			}
			catch (Exception e)
			{
				// Clean things up
				if (id != -1)
					try
					{
						cmd = conn.CreateCommand();
						cmd.CommandText =
@"DELETE FROM scpuzzles WHERE pz_id = " + id.ToString();
						cmd.ExecuteNonQuery();

						cmd = conn.CreateCommand();
						cmd.CommandText =
@"DELETE FROM scpuzzlefrag WHERE pzf_puzzle = " + id.ToString();
						cmd.ExecuteNonQuery();
					}
					catch {}
				throw (e);
			}
			finally
			{
				try
				{
					conn.StateRestore(st);
				}
				catch {}
			}
			return (id);
		}

		/// <summary>
		/// Load a puzzle's data from the database.
		/// </summary>
		/// <param name="pzid">The puzzle's database ID.</param>
		public void LoadPuzzle(int pzid)
		{
			ConnectionState st;
			IDbCommand cmd;
			IDbDataParameter param;
			IDataReader rdr;
			int imgid;
			SCDBImage img;
			SCPuzzleFrag[,] p;
			SCPuzzleFrag f;
			
			if (disabledb)
				throw (new ApplicationException(
					"For operation without a database, use BuildPuzzle()"));

			st = ConnectionState.Closed;
			rdr = null;
			try
			{
				cmd = conn.CreateCommand();
				cmd.CommandText =
@"SELECT * FROM scpuzzles WHERE pz_id = @id";
				param = cmd.CreateParameter();
				param.ParameterName = "@id";
				param.DbType = DbType.Int32;
				param.Value = pzid;
				cmd.Parameters.Add(param);

				st = conn.StateOpen();
				rdr = cmd.ExecuteReader();
				if (rdr == null || !rdr.Read())
					throw (new ApplicationException(
					    "No puzzle with id " + id.ToString() +
					    "found in the database"));
				imgid = (int)rdr["pz_pic"];
				moves = (int)rdr["pz_moves"];
				curx = (int)rdr["pz_curx"];
				cury = (int)rdr["pz_cury"];
				rdr.Close();
				rdr = null;

				img = new SCDBImage();
				img.ConnectionProperties = prop;
				img.LoadImage(imgid);
				p = new SCPuzzleFrag[img.Image.Rows, img.Image.Columns];

				cmd = conn.CreateCommand();
				cmd.CommandText =
@"SELECT * FROM scpuzzlefrag WHERE pzf_puzzle = @id";
				param = cmd.CreateParameter();
				param.ParameterName = "@id";
				param.DbType = DbType.Int32;
				param.Value = pzid;
				cmd.Parameters.Add(param);

				rdr = cmd.ExecuteReader();
				if (rdr == null || !rdr.Read())
					throw (new ApplicationException(
					    "No pieces found for puzzle " +
					    id.ToString()));
				do
				{
					f.x = (int)rdr["pzf_x"];
					f.y = (int)rdr["pzf_y"];
					f.imx = (int)rdr["pzf_imx"];
					f.imy = (int)rdr["pzf_imy"];
					p[f.y, f.x] = f;
				} while (rdr.Read());
				rdr.Close();
				rdr = null;

				id = pzid;
				image = img;
				pieces = p;
			}
			finally
			{
				if (rdr != null)
					try
					{
						rdr.Close();
					}
					catch{}
				try
				{
					conn.StateRestore(st);
				}
				catch {}
			}
		}

		/// <summary>
		/// Determine whether the puzzle has been solved.
		/// </summary>
		/// <returns>True if the puzzle has been solved.</returns>
		public bool IsOrdered()
		{
			uint rows, cols, x, y;

			rows = image.Image.Rows;
			cols = image.Image.Columns;
			for (y = 0; y < rows; y++)
				for (x = 0; x < cols; x++)
					if (pieces[y, x].imx != pieces[y, x].x ||
					    pieces[y, x].imy != pieces[y, x].y)
						return (false);
			return (true);
		}

		/// <summary>
		/// Initialize the puzzle pieces array.
		/// </summary>
		public void InitPieces()
		{
			uint rows, cols, x, y;
			SCPuzzleFrag[,] p;
			
			rows = image.Image.Rows;
			cols = image.Image.Columns;
			p = new SCPuzzleFrag[rows, cols];
			for (y = 0; y < rows; y++)
				for (x = 0; x < cols; x++)
				{
					p[y, x].x = p[y, x].imx = (int)x;
					p[y, x].y = p[y, x].imy = (int)y;
				}
			pieces = p;
		}

		/// <summary>
		/// Load the image of a puzzle fragment.
		/// </summary>
		/// <param name="pzid">The puzzle to use if no currently loaded.</param>
		/// <param name="x">The X coordinate of the fragment.</param>
		/// <param name="y">The Y coordinate of the fragment.</param>
		/// <returns>The image at the specified position.</returns>
		public Image LoadFragmentImage(int pzid, uint x, uint y)
		{
			ConnectionState st;
			IDbCommand cmd;
			IDbDataParameter param;
			FileStream fs;
			Image img, nimg;
			string fname;
			
			if (disabledb)
				throw (new ApplicationException(
					"For operation without a database, use BuildPuzzle()"));

			st = ConnectionState.Closed;
			fs = null;
			img = null;
			try
			{
				cmd = conn.CreateCommand();
				cmd.CommandText =
@"SELECT frag_fname
FROM scpuzzlefrag
JOIN scpuzzles ON pz_id = pzf_puzzle
JOIN scfragments ON frag_pic = pz_pic AND frag_x = pzf_imx AND frag_y = pzf_imy
WHERE pzf_puzzle=@id and pzf_x=@x and pzf_y=@y";

				param = cmd.CreateParameter();
				param.ParameterName = "@id";
				param.DbType = DbType.Int32;
				if (id == -1)
					param.Value = pzid;
				else
					param.Value = id;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.ParameterName = "@x";
				param.DbType = DbType.Int32;
				param.Value = (int)x;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.ParameterName = "@y";
				param.DbType = DbType.Int32;
				param.Value = (int)y;
				cmd.Parameters.Add(param);

				st = conn.StateOpen();
				fname = (string)cmd.ExecuteScalar();
				fs = new FileStream(fname, FileMode.Open);
				nimg = new Bitmap(fs);
				img = new Bitmap(nimg);
				fs.Close();
				fs = null;
			}
			finally
			{
				try
				{
					conn.StateRestore(st);
				}
				catch {}
				if (fs != null)
					try
					{
						fs.Close();
					}
					catch {}
			}
			return (img);
		}

		/// <summary>
		/// Switch the two specified puzzle fragments.
		/// </summary>
		/// <param name="x1">The X coordinate of the first fragment.</param>
		/// <param name="y1">The Y coordinate of the first fragment</param>
		/// <param name="x2">The X coordinate of the second fragment.</param>
		/// <param name="y2">The Y coordinate of the second fragment.</param>
		public void SwitchFragmentsDB(int x1, int y1, int x2, int y2)
		{
			ConnectionState st;
			IDbCommand cmd;
			IDbDataParameter param;
			string str;
			
			st = ConnectionState.Closed;
			try
			{
				str =
@"UPDATE scpuzzlefrag
SET pzf_imx = @imx, pzf_imy= @imy
WHERE pzf_puzzle = @id AND pzf_x = @x AND pzf_y = @y";

				cmd = conn.CreateCommand();
				cmd.CommandText = str;
				param = cmd.CreateParameter();
				param.ParameterName = "@id";
				param.DbType = DbType.Int32;
				param.Value = id;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.ParameterName = "@x";
				param.DbType = DbType.Int32;
				param.Value = x1;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.ParameterName = "@y";
				param.DbType = DbType.Int32;
				param.Value = y1;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.ParameterName = "@imx";
				param.DbType = DbType.Int32;
				param.Value = pieces[y2, x2].imx;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.ParameterName = "@imy";
				param.DbType = DbType.Int32;
				param.Value = pieces[y2, x2].imy;
				cmd.Parameters.Add(param);

				st = conn.StateOpen();
				cmd.ExecuteNonQuery();

				cmd = conn.CreateCommand();
				cmd.CommandText = str;
				param = cmd.CreateParameter();
				param.ParameterName = "@id";
				param.DbType = DbType.Int32;
				param.Value = id;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.ParameterName = "@x";
				param.DbType = DbType.Int32;
				param.Value = x2;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.ParameterName = "@y";
				param.DbType = DbType.Int32;
				param.Value = y2;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.ParameterName = "@imx";
				param.DbType = DbType.Int32;
				param.Value = pieces[y1, x1].imx;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.ParameterName = "@imy";
				param.DbType = DbType.Int32;
				param.Value = pieces[y1, x1].imy;
				cmd.Parameters.Add(param);

				cmd.ExecuteNonQuery();

			}
			finally
			{
				try
				{
					conn.StateRestore(st);
				}
				catch {}
			}
		}

		/// <summary>
		/// Switch the two specified puzzle fragments.
		/// </summary>
		/// <param name="x1">The X coordinate of the first fragment.</param>
		/// <param name="y1">The Y coordinate of the first fragment</param>
		/// <param name="x2">The X coordinate of the second fragment.</param>
		/// <param name="y2">The Y coordinate of the second fragment.</param>
		public void SwitchFragments(int x1, int y1, int x2, int y2)
		{
			int tx, ty;

			if (!disabledb)
				SwitchFragmentsDB(x1, y1, x2, y2);

			tx = pieces[y1, x1].imx;
			ty = pieces[y1, x1].imy;
			pieces[y1, x1].imx = pieces[y2, x2].imx;
			pieces[y1, x1].imy = pieces[y2, x2].imy;
			pieces[y2, x2].imx = tx;
			pieces[y2, x2].imy = ty;
		}

		/// <summary>
		/// Can the current piece move up?
		/// </summary>
		/// <returns>True if the move is possible.</returns>
		public bool CanMoveUp()
		{
			return (cury > 0);
		}

		/// <summary>
		/// Can the current piece move down?
		/// </summary>
		/// <returns>True if the move is possible.</returns>
		public bool CanMoveDown()
		{
			return (cury < image.Image.Rows - 1);
		}

		/// <summary>
		/// Can the current piece move left?
		/// </summary>
		/// <returns>True if the move is possible.</returns>
		public bool CanMoveLeft()
		{
			return (curx > 0);
		}

		/// <summary>
		/// Can the current piece move right?
		/// </summary>
		/// <returns>True if the move is possible.</returns>
		public bool CanMoveRight()
		{
			return (curx < image.Image.Columns - 1);
		}

		/// <summary>
		/// Move the current piece up.
		/// </summary>
		public void MoveUp()
		{
			if (!CanMoveUp())
				return;
			SwitchFragments(curx, cury, curx, cury - 1);
			cury--;
			if (!hold)
				UpdateDBPos();
		}

		/// <summary>
		/// Move the current piece down.
		/// </summary>
		public void MoveDown()
		{
			if (!CanMoveDown())
				return;
			SwitchFragments(curx, cury, curx, cury + 1);
			cury++;
			if (!hold)
				UpdateDBPos();
		}

		/// <summary>
		/// Move the current piece left.
		/// </summary>
		public void MoveLeft()
		{
			if (!CanMoveLeft())
				return;
			SwitchFragments(curx, cury, curx - 1, cury);
			curx--;
			if (!hold)
				UpdateDBPos();
		}

		/// <summary>
		/// Move the current piece right.
		/// </summary>
		public void MoveRight()
		{
			if (!CanMoveRight())
				return;
			SwitchFragments(curx, cury, curx + 1, cury);
			curx++;
			if (!hold)
				UpdateDBPos();
		}

		/// <summary>
		/// Can the current piece move in the given direction?
		/// </summary>
		/// <param name="dir">The direction to test.</param>
		/// <returns>True if the move is possible.</returns>
		public bool CanMove(SCDir dir)
		{
			switch (dir)
			{
				case SCDir.Up:
					return (CanMoveUp());

				case SCDir.Down:
					return (CanMoveDown());

				case SCDir.Left:
					return (CanMoveLeft());

				case SCDir.Right:
					return (CanMoveRight());

				default:
					throw (new ArgumentException(
					    "Invalid direction " + dir,
					    "dir"));
			}
		}

		/// <summary>
		/// Move the current piece in the given direction.
		/// </summary>
		public void Move(SCDir dir)
		{
			switch (dir)
			{
				case SCDir.Up:
					MoveUp();
					break;

				case SCDir.Down:
					MoveDown();
					break;

				case SCDir.Left:
					MoveLeft();
					break;

				case SCDir.Right:
					MoveRight();
					break;

				default:
					throw (new ArgumentException(
					    "Invalid direction " + dir,
					    "dir"));
			}
		}

		/// <summary>
		/// "Scramble" the puzzle in a series of random moves.
		/// </summary>
		/// <param name="movecount">The number of moves.</param>
		public void Scramble(int movecount)
		{
			SCDir dir;
			ArrayList l;
			SCDir[] m;
			Random rnd;
			int curmove;

			if (movecount < 1)
				return;
			l = new ArrayList();
			rnd = new Random();
			
			// The first move is special
			hold = true;
			dir = rnd.Next(2) == 1? SCDir.Up: SCDir.Left;
			Move(dir);
			dir = ReverseDir(dir);

			for (curmove = 1; curmove < movecount; curmove++)
			{
				l.Clear();
				if (dir != SCDir.Up && CanMoveUp())
					l.Add(SCDir.Up);
				if (dir != SCDir.Down && CanMoveDown())
					l.Add(SCDir.Down);
				if (dir != SCDir.Left && CanMoveLeft())
					l.Add(SCDir.Left);
				if (dir != SCDir.Right && CanMoveRight())
					l.Add(SCDir.Right);
				m = (SCDir[])l.ToArray(typeof(SCDir));
				dir = m[rnd.Next(m.Length)];
				Move(dir);
				dir = ReverseDir(dir);
			}

			hold = false;
			UpdateDBPos();
		}

		/// <summary>
		/// Get the direction opposite to the specified one.
		/// </summary>
		/// <param name="dir">The direction to reverse.</param>
		/// <returns>The direction opposite to the specified one.</returns>
		protected SCDir ReverseDir(SCDir dir)
		{
			switch (dir)
			{
				case SCDir.Up:
					return (SCDir.Down);

				case SCDir.Down:
					return (SCDir.Up);

				case SCDir.Left:
					return (SCDir.Right);

				case SCDir.Right:
					return (SCDir.Left);

				default:
					throw (new ArgumentException(
					    "Invalid direction " + dir,
					    "dir"));
			}
		}

		/// <summary>
		/// Update the database's idea of the currently selected piece.
		/// </summary>
		protected void UpdateDBPos()
		{
			ConnectionState st;
			IDbCommand cmd;

			if (disabledb)
				return;

			st = ConnectionState.Closed;
			try
			{
				cmd = conn.CreateCommand();
				cmd.CommandText = String.Format(
@"UPDATE scpuzzles SET pz_curx = {0}, pz_cury = {1} WHERE pz_id = {2}",
				    curx, cury, id);
				st = conn.StateOpen();
				cmd.ExecuteNonQuery();
			}
			finally
			{
				try
				{
					conn.StateRestore(st);
				}
				catch {}
			}
		}

		/// <summary>
		/// Update the database's idea of the number of moves done.
		/// </summary>
		protected void UpdateDBMoves()
		{
			ConnectionState st;
			IDbCommand cmd;

			if (disabledb)
				return;

			st = ConnectionState.Closed;
			try
			{
				cmd = conn.CreateCommand();
				cmd.CommandText = String.Format(
@"UPDATE scpuzzles SET pz_moves = {0} WHERE pz_id = {1}",
					moves, id);
				st = conn.StateOpen();
				cmd.ExecuteNonQuery();
			}
			finally
			{
				try
				{
					conn.StateRestore(st);
				}
				catch {}
			}
		}

		/// <summary>
		/// Get or set the number of moves made so far.
		/// </summary>
		public int Moves
		{
			get
			{
				return (moves);
			}

			set
			{
				moves = value;
				UpdateDBMoves();
			}
		}

		/// <summary>
		/// Get or set the database bypass flag.
		/// </summary>
		public bool DisableDB
		{
			get
			{
				return (disabledb);
			}

			set
			{
				disabledb = value;
			}
		}

		/// <summary>
		/// Build a puzzle from image data obtained somewhere else
		/// (e.g. by Unscramble! via Scramble's PuzzleData webservice)
		/// </summary>
		/// <param name="data">The image data.</param>
		/// <param name="pieces">The image fragments.</param>
		public void BuildPuzzle(SCImageData data,
		    Image[,] pieces)
		{

			image = new SCDBImage();
			image.Image = new SCImage();
			image.Image.Data = data;
			image.Image.Pieces = pieces;
			InitPieces();
			curx = (int)image.Image.Columns - 1;
			cury = (int)image.Image.Rows - 1;
			moves = 0;
		}

		/// <summary>
		/// Get the array of puzzle fragment positions.
		/// </summary>
		public SCPuzzleFrag[,] Pieces
		{
			get
			{
				return (pieces);
			}
		}
	}
}
