using System;
using System.Data;

using Ringlet.Data.ConnectionUtil;
using Ringlet.Misc.VCS;

namespace Ringlet.Drawing.Scramble
{
	/// <summary>
	/// A structure representing a single high score entry.
	/// </summary>
	public struct SCScore : IComparable, IVersioned
	{
		/// <summary>
		/// The database identifier of the high score entry.
		/// </summary>
		public int id;
		/// <summary>
		/// The player name.
		/// </summary>
		public string name;
		/// <summary>
		/// The image name.
		/// </summary>
		public string picture;
		/// <summary>
		/// The number of image rows.
		/// </summary>
		public int rows;
		/// <summary>
		/// The number of image columns.
		/// </summary>
		public int cols;
		/// <summary>
		/// The number of moves the player made.
		/// </summary>
		public int moves;
		/// <summary>
		/// The date/time the score entry was added.
		/// </summary>
		public DateTime date;

		/// <summary>
		/// Initialize a high score object from a data record.
		/// </summary>
		/// <param name="rdr">The reader to get the data from.</param>
		public SCScore(IDataReader rdr)
		{
			id = (int)rdr["sc_id"];
			name = (string)rdr["sc_name"];
			picture = (string)rdr["sc_picture"];
			rows = (int)rdr["sc_rows"];
			cols = (int)rdr["sc_cols"];
			moves = (int)rdr["sc_moves"];
			date = (DateTime)rdr["sc_date"];
		}

		/// <summary>
		/// The version control system tag.
		/// </summary>
		public string VCSTag
		{
			get
			{
				return ("$Ringlet: cs/misc/Scramble/libscpic/SCScore.cs,v 1.1 2003/07/11 18:36:32 roam Exp $");
			}
		}

		/// <summary>
		/// Compare two SCScore objects on number of moves.
		/// </summary>
		/// <param name="obj">The destination object to compare to.</param>
		/// <returns>IComparer.CompareTo() result.</returns>
		public int CompareTo(object obj)
		{
			SCScore s;
			int val;

			s = (SCScore)obj;
			if ((val = moves.CompareTo(s.moves)) != 0)
				return (val);
			else if ((val = (rows * cols).CompareTo(s.rows * s.cols)) != 0)
				return (-val);
			else if ((val = date.CompareTo(s.date)) != 0)
				return (-val);
			else if ((val = picture.CompareTo(s.picture)) != 0)
				return (val);
			else if ((val = name.CompareTo(s.name)) != 0)
				return (val);
			else
				return (id.CompareTo(s.id));
		}

		/// <summary>
		/// Add a high-score object to the database and update
		/// its 'id' field.
		/// </summary>
		/// <param name="conn">The database connection to use.</param>
		public void AddScore(Connection conn)
		{
			ConnectionState st;
			IDbCommand cmd;
			IDbDataParameter param;

			st = ConnectionState.Closed;
			try
			{
				cmd = conn.CreateCommand();
				cmd.CommandText =
@"INSERT INTO scscores(sc_name, sc_picture, sc_rows, sc_cols, sc_moves, sc_date)
VALUES(@name, @pic, @rows, @cols, @moves, @date)";
				
				param = cmd.CreateParameter();
				param.ParameterName = "@name";
				param.DbType = DbType.String;
				param.Value = name;
				cmd.Parameters.Add(param);
				
				param = cmd.CreateParameter();
				param.ParameterName = "@pic";
				param.DbType = DbType.String;
				param.Value = picture;
				cmd.Parameters.Add(param);
				
				param = cmd.CreateParameter();
				param.ParameterName = "@rows";
				param.DbType = DbType.Int32;
				param.Value = rows;
				cmd.Parameters.Add(param);
				
				param = cmd.CreateParameter();
				param.ParameterName = "@cols";
				param.DbType = DbType.Int32;
				param.Value = cols;
				cmd.Parameters.Add(param);
				
				param = cmd.CreateParameter();
				param.ParameterName = "@moves";
				param.DbType = DbType.Int32;
				param.Value = moves;
				cmd.Parameters.Add(param);

				param = cmd.CreateParameter();
				param.ParameterName = "@date";
				param.DbType = DbType.DateTime;
				param.Value = date;
				cmd.Parameters.Add(param);

				st = conn.StateOpen();
				cmd.ExecuteNonQuery();

				id = conn.LastInsertID();
			}
			finally
			{
				try
				{
					conn.StateRestore(st);
				}
				catch {}
			}
		}

		/// <summary>
		/// Remove a high-score object from the database.
		/// </summary>
		/// <param name="conn">The database connection to use.</param>
		public void DeleteScore(Connection conn)
		{
			ConnectionState st;
			IDbCommand cmd;

			st = ConnectionState.Closed;
			try
			{
				cmd = conn.CreateCommand();
				cmd.CommandText = String.Format(
@"DELETE FROM scscores WHERE sc_id = {0}", id);
				st = conn.StateOpen();
				cmd.ExecuteNonQuery();
			}
			finally
			{
				try
				{
					conn.StateRestore(st);
				}
				catch {}
			}
		}
	}
}
